#lang typed/racket

(require "common-types.rkt")

;; Configure the transformer for a specific semantics.
;; The required semantics must be levels-compatible, that is,
;; in addition to the usual interface it must provide the lookup form
(require (only-in "EL-semantics.rkt"
                  M                  
                  M-return
                  M-bind
                  M-do
                  M-make-aspect
                  MakeAspect
                  MakeAspectResult
                  
                  ;; user-visible syntax from the imported semantics
                  run
                  run-debug
                  up
                  down
                  up:
                  down:
                  lookup
                  lambda-s
                  lambda-s:
                  ))

(require/typed "monad.rkt"
               [struct monad
                 ([return : (All (A) (A -> (M A)))]
                  [bind : (All (A B) ((M A) (A -> (M B)) -> (M B)))])])

(define-type LA (All (A) (M (U A Exception))))

(: LA-return (All (A) (A -> (LA A))))
(define (LA-return a)  
  (M-return a))

(: LA-bind (All (A B) ((LA A) (A -> (LA B)) -> (LA B))))
(define (LA-bind mae f)  
  (M-bind mae
          (lambda: ([a : (U A Exception)])
                   (if (Exception? a)
                       (M-return (cons 'Exception (cdr a)))
                       (f a)))))

(: raise* (All (A) (A -> (M Exception))))
(define (raise* a)
  (M-do
   (n <- (lookup) : Integer)
   (M-return (cons 'Exception n))))

(: try-with ((LA Any) (LA Any) -> (LA Any)))
(define (try-with e1 e2)
  (M-do
   (result <- e1 : (U Any Exception))
   (if (Exception? result)
       (M-do
        (n <- (lookup) : Integer)
        (if (= n (cdr result))
            e2
            (M-return (cons 'Exception (cdr result)))))           
       (M-return result))))

(define-syntax-rule (raise*/prim e)
  (#%app raise* e))

(define-syntax-rule (try-with/prim e1 e2)
  (#%app try-with e1 e2))

(define-syntax LA-do
  (syntax-rules (do-with)
    ((LA-do s ...) (do-with LA-bind s ...))))

(provide
 ;; Monascheme interface
 (rename-out [LA M]
             [LA-return M-return]
             [LA-bind M-bind]
             [LA-do M-do])
 M-make-aspect
 MakeAspect
 MakeAspectResult
 
 ;; User-visible syntax
 run
 run-debug
 up
 down
 up:
 down:
 lookup
 lambda-s
 lambda-s:
 (rename-out [raise*/prim raise*]
             [try-with/prim try-with])) 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Macros
;; Each module needs its own macros (with the same code!)
;; because they depend on the concrete types of return and bind
;; that are only known at compile time.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Chains operations using the binder of the current monad.
(define-syntax do
  (syntax-rules (lambda let let* letrec letrec* <-)
    ((do s) s)
    ((do (x <- s) ss ...) ((monad-bind (get-current-monad)) s (lambda (x) (do ss ...))))
    ((do (x <- s : T) ss ...) ((monad-bind (get-current-monad)) s (lambda: ([x : T]) (do ss ...))))
    ((do (let bs) ss ...) (let bs (do ss ...)))
    ((do (let* bs) ss ...) (let* bs (do ss ...)))
    ((do (letrec bs) ss ...) (letrec bs (do ss ...)))
    ((do (letrec* bs) ss ...) (letrec* bs (do ss ...)))
    ((do s ss ...) ((monad-bind (get-current-monad)) s (lambda (_) (do ss ...))))))

;; Chains operations using a specific binder B.
(define-syntax do-with
  (syntax-rules (lambda let let* letrec letrec* <-)
    ((do-with B s) s)
    ((do-with B (x <- s) ss ...) (B s (lambda (x) (do-with B ss ...))))
    ((do-with B (x <- s : T) ss ...) (B s (lambda: ([x : T]) (do-with B ss ...))))
    ((do-with B (let bs) ss ...) (let bs (do-with B ss ...)))
    ((do-with B (let* bs) ss ...) (let* bs (do-with B ss ...)))
    ((do-with B (letrec bs) ss ...) (letrec bs (do-with B ss ...)))
    ((do-with B (letrec* bs) ss ...) (letrec* bs (do-with B ss ...)))
    ((do-with B s ss ...) (B s (lambda (_) (do-with B ss ...))))))

;; Applies Racket primitive functions with computations as arguments,
;; and provides a computation as a result. Uses the current monad.
(define-syntax (fmap stx)
  (syntax-case stx (do <-)
    ((fmap f a ...)
     (with-syntax ([(y ...) (generate-temporaries #'(a ...))])
       #'(do (y <- a) ...
             ((monad-return (get-current-monad)) (f y ...)))))))