#lang typed/racket

(require "common-types.rkt")

;; Structure to carry monad information
(struct: (A) ValueLevel ([value : A] [level : Integer]) #:transparent)

;; Function type for the execution level monad,
;; which is a special case of the State monad.
(struct: (A) EL ([value : (Integer -> (ValueLevel A))]) #:transparent)

(require/typed "monad.rkt"
               [struct monad
                 ([return : (All (A) (A -> (EL A)))]
                  [bind : (All (A B) ((EL A) (A -> (EL B)) -> (EL B)))])]
               [current-monad (Parameterof (U monad False))])

(: get-current-monad (-> monad))
(define (get-current-monad)
  (let ((M (current-monad)))
    (cond
      ((monad? M) M)
      (else (error "Current monad has not been specified")))))

;; Monadic values are functions Level -> (a . Level)
;; In this case Level = Int
(: EL-return (All (A) (A -> (EL A))))
(define (EL-return a)
  (EL
   (λ: ([s0 : Integer])
       (ValueLevel a s0))))

(: EL-bind (All (A B) ((EL A) (A -> (EL B)) -> (EL B))))
(define (EL-bind ma f)
  (EL
   (λ: ([s0 : Integer])                       
       (let ([result ((EL-value ma) s0)])
         ((EL-value (f (ValueLevel-value result)))
          (ValueLevel-level result))))))

;; Access the current execution level
(: lookup (-> (EL Integer)))
(define (lookup)
  (EL
   (λ: ([s : Integer])
       ((EL-value (EL-return s)) s))))

;; Increase execution level by one
(: inc-level (-> (EL Void)))
(define (inc-level)
  (EL
   (λ: ([s : Integer])
       (ValueLevel (void) (+ s 1)))))


;; Decrease execution level by one
(: dec-level (-> (EL Void)))
(define (dec-level)
  (EL
   (λ: ([s : Integer])
       (ValueLevel (void) (- s 1)))))

;; Shifts execution level to l
(: shift (Integer -> (EL Void)))
(define (shift l)
  (EL
   (λ: ([s : Integer])
       (ValueLevel (void) l))))

;; Runs a program with initial level 0
(: run (All (A) ((EL A) -> A)))
(define (run t) (ValueLevel-value ((EL-value t) 0)))

;; For debugging, gives the value and the level
(: run-debug (All (A) ((EL A) -> (ValueLevel A))))
(define (run-debug t) ((EL-value t) 0))

(define-type MakeAspectResult (Pairof (JPStack -> (EL Any))
                                      ((Any -> (EL Any)) -> (Any -> (EL Any)))))

(define-type MakeAspect
  ((JPStack -> (EL Any)) ((Any -> (EL Any)) -> (Any -> (EL Any))) -> (EL MakeAspectResult)))

(: make-asp MakeAspect)   
(define (make-asp pc adv)
  (do
   (n <- (lookup) : Integer)
   ((monad-return (get-current-monad))
    (cons (λ: ([jp : JPStack])
              (do               
               (l <- (lookup) : Integer)
               (if (= l n)
                   (up (pc jp))
                   ((monad-return (get-current-monad)) #f))))              
          
          (λ: ([p : (Any -> (EL Any))])             
              (λ: ([arg : Any])                 
                  (do                      
                   (up ((adv (lambda-at n (args) (p args))) arg)))))))))

;; Avoids weaving on run
(define-syntax-rule (run/prim t)
  (#%app run t))

(define-syntax-rule (run-debug/prim t)
  (#%app run-debug t))

;; Executes e one level above
(define-syntax-rule (up e)
  (EL-do
   (inc-level)       
   (x <- e : Any)
   (dec-level)
   (EL-return x)))

;; Executes e one level above, and annotates it with type T
(define-syntax-rule (up: e : T)
  (EL-do
   (inc-level)       
   (x <- e : T)
   (dec-level)
   (EL-return x)))

;; Executes e one level below
(define-syntax-rule (down e)
  (EL-do
   (dec-level)
   (x <- e : Any)
   (inc-level)
   (EL-return x)))  

;; Executes e one level below, and annotates it with type T
(define-syntax-rule (down: e : T)
  (EL-do
   (dec-level)
   (x <- e : T)
   (inc-level)
   (EL-return x)))  

;; Executes e at level l
(define-syntax-rule (at l e)
  (EL-do
   (current <- (lookup) : Integer)
   (shift l)      
   (x <- e : Any)
   (shift current)
   (EL-return x)))

;; Executes e at level l, and annotates it with type T
(define-syntax-rule (at: l e : T)
  (EL-do
   (current <- (lookup) : Integer)     
   (shift l)
   (x <- e : T)
   (shift current)
   (EL-return x)))

;; Instantiates a level-capturing function at a given level
(define-syntax-rule (lambda-at level params body ...)     
  (lambda params (at level (EL-do body ...))))

(define-syntax-rule (lambda-at: level params body ...)     
  (lambda: params (at level (EL-do body ...))))

;; Instantiates a level-capturing function, using the current level
(define-syntax-rule (lambda-s params body ...)
  (EL-do
   (c <- (lookup) : Integer)
   (EL-return (lambda-at c params body ...))))

(define-syntax-rule (lambda-s: params body ...)
  (EL-do
   (c <- (lookup) : Integer)
   (EL-return (lambda-at: c params body ...))))

(define-syntax EL-do
  (syntax-rules (do-with)
    ((EL-do s ...) (do-with EL-bind s ...))))

(provide
 ;; Monad configuration
 (rename-out [EL M]
             [EL-return M-return]
             [EL-bind M-bind]
             [EL-do M-do]
             [make-asp M-make-aspect])
 MakeAspect
 MakeAspectResult
 
 ;; User-visible syntax
 (rename-out [run/prim run]
             [run-debug/prim run-debug])
 up
 down
 up:
 down:
 lookup
 lambda-s
 lambda-s:) 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Macros
;; Each module needs its own macros (with the same code!)
;; because they depend on the concrete types of return and bind
;; that are only known at compile time.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Chains operations using the binder of the current monad.
(define-syntax do
  (syntax-rules (lambda let let* letrec letrec* <-)
    ((do s) s)
    ((do (x <- s) ss ...) ((monad-bind (get-current-monad)) s (lambda (x) (do ss ...))))
    ((do (x <- s : T) ss ...) ((monad-bind (get-current-monad)) s (lambda: ([x : T]) (do ss ...))))
    ((do (let bs) ss ...) (let bs (do ss ...)))
    ((do (let* bs) ss ...) (let* bs (do ss ...)))
    ((do (letrec bs) ss ...) (letrec bs (do ss ...)))
    ((do (letrec* bs) ss ...) (letrec* bs (do ss ...)))
    ((do s ss ...) ((monad-bind (get-current-monad)) s (lambda (_) (do ss ...))))))

;; Chains operations using a specific binder B.
(define-syntax do-with
  (syntax-rules (lambda let let* letrec letrec* <-)
    ((do-with B s) s)
    ((do-with B (x <- s) ss ...) (B s (lambda (x) (do-with B ss ...))))
    ((do-with B (x <- s : T) ss ...) (B s (lambda: ([x : T]) (do-with B ss ...))))
    ((do-with B (let bs) ss ...) (let bs (do-with B ss ...)))
    ((do-with B (let* bs) ss ...) (let* bs (do-with B ss ...)))
    ((do-with B (letrec bs) ss ...) (letrec bs (do-with B ss ...)))
    ((do-with B (letrec* bs) ss ...) (letrec* bs (do-with B ss ...)))
    ((do-with B s ss ...) (B s (lambda (_) (do-with B ss ...))))))

;; Applies Racket primitive functions with computations as arguments,
;; and provides a computation as a result. Uses the current monad.
(define-syntax (fmap stx)
  (syntax-case stx (do <-)
    ((fmap f a ...)
     (with-syntax ([(y ...) (generate-temporaries #'(a ...))])
       #'(do (y <- a) ...
             ((monad-return (get-current-monad)) (f y ...)))))))