#lang typed/racket
(require "common-types.rkt")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Monascheme Configuration
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(require (only-in
          ;"EL-semantics.rkt"
          "EL-LA-semantics.rkt"
          M          
          M-return
          M-bind
          M-do
          M-make-aspect
          MakeAspect
          MakeAspectResult          
          ;; new user-visible syntax, this is configurable
          run run-debug up down up: down: lookup lambda-s lambda-s:
          ;; la-exceptions
          raise*
          try-with
          ))

;; Provide the type name to be used by user programs
;; and the additional user-visible syntax
(provide (rename-out [M EL])
         ;; this part is also configurable
         run run-debug up down up: down: lookup lambda-s lambda-s:
         ;; la-exceptions
         raise*
         try-with
         )

;; The type of these identifiers depends on the M type constructor.
;; Currently this is the only way to do it in Typed Racket.
(require/typed "monad.rkt"               
               [struct monad
                 ([return : (All (A) (A -> (M A)))]
                  [bind : (All (A B) ((M A) (A -> (M B)) -> (M B)))])]
               [current-monad (Parameterof (U monad False))]
               [current-makeasp (Parameterof (U MakeAspect False))])

;; We need to configure the parameters so this module typechecks
(begin
  (current-monad (monad M-return M-bind))
  (current-makeasp M-make-aspect))  

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Parameters accessors
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(: get-current-monad (-> monad))
(define (get-current-monad)
  (let ((m (current-monad)))
    (cond
      ((monad? m)m )
      (else (error "Current monad has not been specified")))))

(: get-current-makeasp (-> MakeAspect))
(define (get-current-makeasp)
  (let ((m (current-makeasp)))
    (cond
      ((not (false? m)) m)
      (else (error "Current makeasp has not been specified")))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Aspect Environment
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(struct: Aspect ([pc : (JPStack -> (M Any))]
                 [adv : ((Any -> (M Any)) -> (Any -> (M Any)))]))

(struct: Undeployer ([asp : Aspect]))

(: dynamic-aspects (Listof Aspect))
(define dynamic-aspects '())

(: static-aspects (Listof Aspect))
(define static-aspects '())

(: top-aspects (Listof Aspect))
(define top-aspects '())

(: get-dynamic-aspects (-> (Listof Aspect)))
(define (get-dynamic-aspects) dynamic-aspects)

(: get-static-aspects (-> (Listof Aspect)))
(define (get-static-aspects) static-aspects)

(: get-top-aspects (-> (Listof Aspect)))
(define (get-top-aspects) top-aspects)

(: add-dynamic-aspect! (Aspect -> (M Undeployer)))
(define (add-dynamic-aspect! asp)
  (set! dynamic-aspects (cons asp (get-dynamic-aspects)))
  (M-return (Undeployer asp)))

(: add-static-aspect! (Aspect -> (M Undeployer)))
(define (add-static-aspect! asp)
  (set! static-aspects (cons asp (get-static-aspects)))
  (M-return (Undeployer asp)))

(: set-static-aspects! ((Listof Aspect) -> (M Void)))
(define (set-static-aspects! aspects)
  (return (set! static-aspects aspects)))

(: add-top-aspect! (Aspect -> (M Undeployer)))
(define (add-top-aspect! asp)  
  (set! top-aspects (cons asp (get-top-aspects)))  
  (M-return (Undeployer asp)))

(: remove-dynamic-aspect! (Undeployer -> (M Void)))
(define (remove-dynamic-aspect! undep)
  (M-return
   (set! dynamic-aspects (filter (negate (λ (x) (eq? x (Undeployer-asp undep)))) dynamic-aspects))))

(: remove-static-aspect! (Undeployer -> (M Void)))
(define (remove-static-aspect! undep)
  (M-return
   (set! static-aspects (filter (negate (λ (x) (eq? x (Undeployer-asp undep)))) static-aspects))))

(: remove-top-aspect! (Undeployer -> (M Void)))
(define (remove-top-aspect! undep)
  (M-return
   (set! top-aspects (filter (negate (λ (x) (eq? x (Undeployer-asp undep)))) top-aspects))))

(: current-aspects (-> (Listof Aspect)))
(define (current-aspects)
  (append (get-dynamic-aspects)
          (get-static-aspects)
          (get-top-aspects)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Join point stack management
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(: joinpoint-stack JPStack)
(define joinpoint-stack '())

(: jp-context (-> JPStack))
(define (jp-context)  
  joinpoint-stack)

(define-syntax (with-joinpoint stx)
  (syntax-case stx ()
    [(_ jp body)
     (syntax
      (do
       (return (set! joinpoint-stack (cons jp joinpoint-stack)))
       (result <- body)
       (return (set! joinpoint-stack (cdr joinpoint-stack)))
       (return result)))]))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Syntax rewriting and support for monadic style
;; Function calls are replaced by calls to app/weave.
;; Definitions using lambda are replaced by lambda/aspects to capture static aspects.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define-syntax (app/weave stx)
  (syntax-case stx ()
    [(_ f a ...)
     (syntax (#%app app/weave/rt f a ...))]))

;; all functions capture and then restore their static aspects
(define-syntax lambda/aspects
  (lambda (stx)
    (syntax-case stx ()
      [(_ params body ...)
       (syntax/loc stx
         (let ([captured-aspects (get-static-aspects)])
           (lambda params
             (do
              (current-static-aspects <- (return (get-static-aspects)) : (Listof Aspect))
              (return (set-static-aspects! captured-aspects))
              (result <- body ...)
              (return (set-static-aspects! current-static-aspects))
              (return result)))))])))

(define-syntax lambda:/aspects
  (lambda (stx)
    (syntax-case stx ()
      [(_ params body ...)
       (syntax/loc stx
         (let ([captured-aspects (get-static-aspects)])
           (lambda: params
                    (do
                     (current-static-aspects <- (return (get-static-aspects)) : (Listof Aspect))
                     (return (set-static-aspects! captured-aspects))
                     (result <- body ...)
                     (return (set-static-aspects! current-static-aspects))
                     (return result)))))])))

;; Simplifies use of return, so it is never woven.
;; Also, return is parameterized by the current-monad
(define-syntax-rule (return e)
  (#%app (#%app monad-return (#%app get-current-monad)) e))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Weaving
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define-type UntypedAppWeaveRtType
  (Procedure Any Procedure Procedure Procedure Procedure -> (M Any)))

(define-type AppWeaveRtType
  (All (A B) ((A -> (M B)) A -> (M Any))))

(require/typed "untyped-weave.rkt"
               [untyped-app/weave/rt UntypedAppWeaveRtType])


(: app/weave/rt ((Nothing -> (M Any)) Any -> (M Any)))
(define (app/weave/rt fun-val arg-vals)
  (with-joinpoint
   (JP (ann fun-val (Nothing -> (M Any))) 'call)
   (begin
     (let ((result (untyped-app/weave/rt fun-val
                                         arg-vals
                                         ;; context for untyped module
                                         jp-context
                                         current-aspects
                                         Aspect-pc
                                         Aspect-adv)))
       result))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Pointcuts - fundamental
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define-syntax-rule (call/prim e)
  (#%app call e))

(: call (Any -> (M (JPStack -> (M Any)))))
(define (call f)
  (return
   (lambda: ([jp* : JPStack])
            (return (eq? f (JP-fun (car jp*)))))))

(define-syntax-rule (top/prim)
  (#%app top))

(: top (-> (M (JPStack -> (M Any)))))
(define (top)
  (return
   (lambda: ([jp* : JPStack])
            (return
             (empty? jp*)))))

;(define (top)
;  (return
;   (lambda: ([jp* : JPStack])
;            (return (null? jp*)))))

(define-syntax-rule (below/prim e)
  (#%app below e))

(: below ((M (JPStack -> (M Any))) -> (M (JPStack -> (M Any)))))
(define (below pc)
  (return
   (lambda: ([jp* : JPStack])
            (do
             (pc-fun <- pc : (JPStack -> (M Any)))
             (return
              (and (not (empty? (rest jp*)))
                   (pc-fun (rest jp*))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Pointcuts - strict combinators
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define-syntax-rule (&&/prim e ...)
  (#%app && e ...))

(: && ((M (JPStack -> (M Any))) * -> (M (JPStack -> (M Any)))))
(define (&& . pcs)
  (return
   (lambda: ([jp* : JPStack])
            (foldr
             (lambda: ([b : (M Any)] [r : (M Any)])                      
                      (do
                       (rr <- r : Any)
                       (bb <- b : Any)
                       (return (and bb rr))))
             (return #t)             
             (map (lambda: ([pc : (M (JPStack -> (M Any)))])
                           (do
                            (pc-fun <- pc : (JPStack -> (M Any)))
                            (pc-fun jp*)))
                  pcs)))))

(define-syntax-rule (||/prim e ...)
  (#%app || e ...))

(: || ((M (JPStack -> (M Any))) * -> (M (JPStack -> (M Any)))))
(define (|| . pcs)
  (return
   (lambda: ([jp* : JPStack])
            (foldr
             (lambda: ([b : (M Any)] [r : (M Any)])                      
                      (do
                       (rr <- r : Any)
                       (bb <- b : Any)
                       (return (or bb rr))))
             (return #t)             
             (map (lambda: ([pc : (M (JPStack -> (M Any)))])
                           (do
                            (pc-fun <- pc : (JPStack -> (M Any)))
                            (pc-fun jp*)))
                  pcs)))))

(define-syntax-rule (!/prim e)
  (#%app ! e))

(: ! ((M (JPStack -> (M Any))) -> (M (JPStack -> (M Any)))))
(define (! pc)
  (do 
   (pc-fun <- pc : (JPStack -> (M Any)))
   (return
    (lambda: ([jp* : JPStack])
             (do 
              (pc-result <- (pc-fun jp*) : Any)
              (return (not pc-result)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Pointcuts - higher-order points-free
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define-syntax-rule (within/prim e)
  (#%app within e))

(: within (Any -> (M (JPStack -> (M Any)))))
(define (within f)
  (below (call f)))

;(define-syntax-rule (cflow/prim e)
;  (#%app cflow e))
;
;(: cflow ((M (JPStack -> (M Any))) -> (M (JPStack -> (M Any)))))
;(define (cflow pc)
;  (&& (! (top))
;      (|| pc
;          (cflowbelow pc))))
;
;(define-syntax-rule (cflowbelow/prim e)
;  (#%app cflowbelow e))
;
;(: cflowbelow ((M (JPStack -> (M Any))) -> (M (JPStack -> (M Any)))))
;(define (cflowbelow pc)
;  (&& (! (top))
;      (below (cflow pc))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Scoped Aspect deployment
;;
;; NOTE: We cannot use parameters for the
;; aspect environment because the monadic translation
;; transforms a program into a closure.
;; Since the macro is expanded at compile time, the
;; dynamic binding would not be captured by the resulting closure
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define-syntax (deploy-fluid stx)
  (syntax-case stx ()
    [(_ pc adv body ...)
     (quasisyntax/loc stx
       (do        
        (v-pc <- pc : (JPStack -> (M Any)))
        (v-adv <- adv : ((Any -> (M Any)) -> (Any -> (M Any))))        
        (asp-pair <- ((get-current-makeasp) v-pc v-adv) : MakeAspectResult)
        (aspect <- (return (Aspect (car asp-pair) (cdr asp-pair))) : Aspect)
        (handle <- (add-dynamic-aspect! aspect) : Undeployer)
        (result <- (do body ...))
        (remove-dynamic-aspect! handle)
        (return result)))]))

(define-syntax (deploy-static stx)
  (syntax-case stx ()
    [(_ pc adv body ...)
     (quasisyntax/loc stx
       (do        
        (v-pc <- pc : (JPStack -> (M Any)))
        (v-adv <- adv : ((Any -> (M Any)) -> (Any -> (M Any))))        
        (asp-pair <- ((get-current-makeasp) v-pc v-adv) : MakeAspectResult)
        (aspect <- (return (Aspect (car asp-pair) (cdr asp-pair))) : Aspect)
        (handle <- (add-static-aspect! aspect) : Undeployer)
        (result <- (do body ...))
        (remove-static-aspect! handle)
        (return result)))]))

(define-syntax (deploy-fluid: stx)  
  (syntax-case stx ()
    [(_ pc adv body ... : T)
     (quasisyntax/loc stx
       (do        
        (v-pc <- pc : (JPStack -> (M Any)))
        (v-adv <- adv : ((Any -> (M Any)) -> (Any -> (M Any))))        
        (asp-pair <- ((get-current-makeasp) v-pc v-adv) : MakeAspectResult)
        (aspect <- (return (Aspect (car asp-pair) (cdr asp-pair))) : Aspect)
        (handle <- (add-dynamic-aspect! aspect) : Undeployer)
        (result <- (do body ...) : T)
        (remove-dynamic-aspect! handle)
        (return result)))]))

(define-syntax (deploy-static: stx)
  (syntax-case stx ()
    [(_ pc adv body ... : T)
     (quasisyntax/loc stx
       (do        
        (v-pc <- pc : (JPStack -> (M Any)))
        (v-adv <- adv : ((Any -> (M Any)) -> (Any -> (M Any))))        
        (asp-pair <- ((get-current-makeasp) v-pc v-adv) : MakeAspectResult)
        (aspect <- (return (Aspect (car asp-pair) (cdr asp-pair))) : Aspect)
        (handle <- (add-static-aspect! aspect) : Undeployer)
        (result <- (do body ...) : T)
        (remove-static-aspect! handle)
        (return result)))]))

;;;;;;;;;;;;;;;;;;;;;;
;; Global deployment
;;;;;;;;;;;;;;;;;;;;;;
(define-syntax-rule (deploy-top pc adv)
  (do        
   (v-pc <- pc : (JPStack -> (M Any)))
   (v-adv <- adv : ((Any -> (M Any)) -> (Any -> (M Any))))        
   (asp-pair <- ((get-current-makeasp) v-pc v-adv) : MakeAspectResult)
   (aspect <- (return (Aspect (car asp-pair) (cdr asp-pair))) : Aspect)
   (handle <- (add-top-aspect! aspect) : Undeployer)
   (return handle)))

(define-syntax-rule (undeploy-top undep)
  (do
   (handle <- undep : Undeployer)
   (remove-top-aspect! handle)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  
;; Exports to be used by client programs
(provide (rename-out
          [#%app app/prim]
          [app/weave #%app]
          [lambda/aspects lambda]
          [lambda:/aspects lambda:])
         
         ;; Types
         JPStack
         
         ;; Parameter manipulation
         current-monad
         current-makeasp
         get-current-monad
         get-current-makeasp
         
         ;; User-visible syntax
         deploy-fluid
         deploy-static
         deploy-top
         undeploy-top
         deploy-fluid:
         deploy-static:
         do
         return        
         fmap
         
         ; pointcuts
         (rename-out [call/prim call]
                     [top/prim top]
                     [below/prim below]   
                     [&&/prim &&]
                     [||/prim ||]
                     [!/prim !]                                                            
                     [within/prim within]
                     ;[cflow/prim cflow]
                     ;[cflowbelow/prim cflowbelow]
                     ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Macros
;; Each module needs its own macros (with the same code!)
;; because they depend on the concrete types of return and bind
;; that are only known at compile time.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Chains operations using the binder of the current monad.
(define-syntax do
  (syntax-rules (lambda let let* letrec letrec* <-)
    ((do s) s)
    ((do (x <- s) ss ...) ((monad-bind (get-current-monad)) s (lambda (x) (do ss ...))))
    ((do (x <- s : T) ss ...) ((monad-bind (get-current-monad)) s (lambda: ([x : T]) (do ss ...))))
    ((do (let bs) ss ...) (let bs (do ss ...)))
    ((do (let* bs) ss ...) (let* bs (do ss ...)))
    ((do (letrec bs) ss ...) (letrec bs (do ss ...)))
    ((do (letrec* bs) ss ...) (letrec* bs (do ss ...)))
    ((do s ss ...) ((monad-bind (get-current-monad)) s (lambda (_) (do ss ...))))))

;; Chains operations using a specific binder B.
(define-syntax do-with
  (syntax-rules (lambda let let* letrec letrec* <-)
    ((do-with B s) s)
    ((do-with B (x <- s) ss ...) (B s (lambda (x) (do-with B ss ...))))
    ((do-with B (x <- s : T) ss ...) (B s (lambda: ([x : T]) (do-with B ss ...))))
    ((do-with B (let bs) ss ...) (let bs (do-with B ss ...)))
    ((do-with B (let* bs) ss ...) (let* bs (do-with B ss ...)))
    ((do-with B (letrec bs) ss ...) (letrec bs (do-with B ss ...)))
    ((do-with B (letrec* bs) ss ...) (letrec* bs (do-with B ss ...)))
    ((do-with B s ss ...) (B s (lambda (_) (do-with B ss ...))))))

;; Applies Racket primitive functions with computations as arguments,
;; and provides a computation as a result. Uses the current monad.
(define-syntax (fmap stx)
  (syntax-case stx (do <-)
    ((fmap f a ...)
     (with-syntax ([(y ...) (generate-temporaries #'(a ...))])
       #'(do (y <- a) ...
             ((monad-return (get-current-monad)) (f y ...)))))))
