#lang typed/racket

(require "monascheme-core.rkt")
(require "test.rkt")
(require "common-types.rkt")

(provide run-test-common)

(: lifted_write (Any -> (EL Any)))
(define lifted_write
  (lambda: ([s : Any]) (fmap write (return s))))

(: lifted_display (Any -> (EL Any)))
(define lifted_display  (lambda: ([a : Any]) (fmap display (return a))))

(define g (lambda (a) (lifted_write a)))
(define f (lambda (a) (g a)))

;; Trace advice
(: trace ((Any -> (EL Any)) -> (Any -> (EL Any))))
(define (trace proceed)  
  (lambda: ([a : Any])  
           (do
            (lifted_write 'a)     
            (proceed a))))

(: top-trace ((Any -> (EL Any)) -> (Any -> (EL Any))))
(define (top-trace proceed)  
  (lambda: ([a : Any])  
           (do
            (lifted_write 'top)     
            (proceed a))))

(: static-trace ((Any -> (EL Any)) -> (Any -> (EL Any))))
(define (static-trace proceed)  
  (lambda: ([a : Any])  
           (do
            (lifted_write 'static)     
            (proceed a))))

(define (run-test-common)  
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;; Fluid/Static deployment
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  
  ;; call is lexically in scope
  (test/str
   (run (deploy-static (call lifted_write)
                       (return trace)
                       (lifted_write 'hello)))
   "ahello")
  
  ;; call is inside the scope of deploy-fluid 
  (test/str
   (run (deploy-fluid (call lifted_write)
                      (return trace)
                      (do
                       (g <- (return (lambda: ([a : Any]) (lifted_write a))) : (Any -> (EL Any)))
                       (g 'hello))))
   "ahello")
  
  ;; call is lexically in scope, inside function g
  (test/str
   (run (deploy-static (call lifted_write)
                       (return trace)
                       (do
                        (g <- (return (lambda: ([a : Any]) (lifted_write a))) : (Any -> (EL Any)))
                        (g 'hello))))
   "ahello")
  
  ;; closure x doesn't capture the dynamically scoped aspect
  (test/str
   (run (do       
         (x <- (deploy-fluid: (call lifted_write)
                              (return trace)
                              (do
                               (return
                                (lambda: ([a : Any]) (lifted_write a))))
                              : (Any -> (EL Any))) : (Any -> (EL Any)))
         (x 'hello)))
   "hello")
  
  ;; closure x captures the statically deployed aspect
  (test/str
   (run (do       
         (x <- (deploy-static: (call lifted_write)
                               (return trace)
                               (do
                                (return
                                 (lambda: ([a : Any]) (lifted_write a))))
                               : (Any -> (EL Any))) : (Any -> (EL Any)))
         (x 'hello)))
   "ahello")
  
  ;; dynamic aspect applies to existing closure in scope
  (test/str
   (run (do
         (x <- (return (lambda: ([a : Any]) (lifted_write a))) : (Any -> (EL Any)))
         (deploy-fluid (call lifted_write) (return trace)
                       (x 'hello))))
   "ahello")
  
  ;; static aspect does not apply to existing closure in scope
  (test/str
   (run (do
         (x <- (return (lambda: ([a : Any]) (lifted_write a))) : (Any -> (EL Any)))
         (deploy-static (call lifted_write) (return trace)
                        (x 'hello))))
   "hello")
  
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;; Global deploymen
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  
  ;; You must evaluate the computation given by deploy-top to deploy the aspect
  (define h1 (run (deploy-top (call lifted_write) (return top-trace))))
  
  (test/str
   (run (lifted_write 'hello))
   "tophello")
  
  (test/str
   (run (deploy-fluid (call lifted_write) (return trace)
                      (lifted_write 'hello)))
   "atophello")
  
  (test/str
   (run (deploy-static (call lifted_write) (return static-trace)
                       (lifted_write 'hello)))
   "statictophello")
  
  (test/str
   (run (deploy-fluid (call lifted_write) (return trace)
                      (deploy-static (call lifted_write) (return static-trace)
                                     (lifted_write 'hello))))
   "astatictophello")
  
  ;; undeploy top aspect: you must evaluate the computation
  (run (undeploy-top (return h1)))
  
  (test/str
   (run (lifted_write 'hello))
   "hello")
  
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;; Execution levels
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  
  ;; level 0
  (test/str
   (run 
    (deploy-fluid (call lifted_write) (return trace)                    
                  (lifted_write 'hello)))
   "ahello")
  
  (test/str
   (run 
    (deploy-fluid (&&               
                   (call lifted_write)               
                   (return (lambda: ([jp : JPStack])
                                    (do
                                     (lifted_write 'pc)                                      
                                     (return #t)))))
                  (return trace)
                  (lifted_write 'hello)))
   "pcahello")
  
  ;; test 1
  (test/str
   (run 
    (up (deploy-fluid (call lifted_write) (return trace)
                      (lifted_write 'hello))))
   "ahello")
  
  (test/str
   (run 
    (up (deploy-fluid (call lifted_write) (return trace)
                      (down (lifted_write 'hello)))))
   "hello")
  
  (test/str
   (run 
    (down (deploy-fluid (call lifted_write) (return trace)
                        (up (lifted_write 'hello)))))
   "hello")  
  
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;; Level-capturing functions
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  (test/equal
   (run (do (f <- (down: (lambda-s: ([a : Any]) (app/prim lookup))
                         : (Any -> (EL Any))) : (Any -> (EL Any)))
            (f 1)))
   -1)
  
  (test/equal
   (run (do (f <- (lambda-s: ([a : Any]) (app/prim lookup)) : (Any -> (EL Any)))
            (f 1)))
   0)
  
  (test/equal
   (run (do (f <- (up: (lambda-s: ([a : Any]) (app/prim lookup))
                       : (Any -> (EL Any))) : (Any -> (EL Any)))
            (f 1)))
   1)
  
  (test/equal
   (run (do (f <- (return (lambda: ([a : Any]) (app/prim lookup))) : (Any -> (EL Any)))
            (f 1)))
   0)  
  
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;; Pointcuts
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  
  ; || 
  (test/str
   (run (deploy-fluid (|| (call lifted_write) (return (lambda: ([jp* : JPStack]) (return #t))))
                      (return trace)
                      (lifted_write 'hello)))
   "ahello")
  
  (test/str
   (run (deploy-fluid (|| (call lifted_write) (call lifted_display))
                      (return trace)
                      (do
                       (lifted_display 'hello)
                       (lifted_write 'world))))
   "ahelloaworld")
  
  ; below
  (test/str
   (run (deploy-fluid (&& (call lifted_write) (below (call g))) (return trace)
                      (g 'hello)))
   "ahello")
  
  (test/str
   (run (deploy-fluid (&& (call lifted_write) (below (call g))) (return trace)
                      (f 'hello)))
   "ahello")
  
  ; within 
  (test/str
   (run (deploy-fluid (&& (call lifted_write) (within g))
                      (return trace)
                      (lifted_write 'hello)))
   "hello")
  
  (test/str
   (run (deploy-fluid (&& (call lifted_write) (within g))
                      (return trace)
                      (g 'hello)))
   "ahello")
  
  (test/str
   (run (deploy-fluid (&& (call lifted_write) (within f)) (return trace)
                      (f 'hello)))
   "ahello")
  
  ;cflow 
  ;(test/str
  ; (run (deploy-fluid (&& (call lifted_write) (cflow (call f))) (return trace)
  ;                    (f 'hello)))
  ; "ahello")
  
  ;cflowbelow
  ;(test/str
  ; (run (deploy-fluid (&& (call lifted_write) (cflowbelow (call f))) (return trace)
  ;                    (f 'hello)))
  ; "ahello")  
  )